from __future__ import annotations

import json
import logging
from collections import OrderedDict, defaultdict
from typing import TYPE_CHECKING

from deptry.rust import get_imports_from_ipynb_files, get_imports_from_py_files

if TYPE_CHECKING:
    from pathlib import Path

    from deptry.rust import Location as RustLocation

from deptry.imports.location import Location


def get_imported_modules_from_list_of_files(list_of_files: list[Path]) -> dict[str, list[Location]]:
    logging.info("Scanning %d %s...", len(list_of_files), "files" if len(list_of_files) > 1 else "file")

    py_files = [str(file) for file in list_of_files if file.suffix == ".py"]
    ipynb_files = [str(file) for file in list_of_files if file.suffix == ".ipynb"]

    modules: dict[str, list[Location]] = defaultdict(list)

    # Process all .py files in parallel using Rust
    if py_files:
        rust_result = get_imports_from_py_files(py_files)
        for module, locations in _convert_rust_locations_to_python_locations(rust_result).items():
            modules[module].extend(locations)

    # Process all .ipynb files in parallel using Rust
    if ipynb_files:
        rust_result = get_imports_from_ipynb_files(ipynb_files)
        for module, locations in _convert_rust_locations_to_python_locations(rust_result).items():
            modules[module].extend(locations)

    sorted_modules = OrderedDict(sorted(modules.items()))
    _log_modules_with_locations(sorted_modules)
    return sorted_modules


def _log_modules_with_locations(modules: dict[str, list[Location]]) -> None:
    modules_dict = {
        module_name: [str(location) for location in locations] for module_name, locations in modules.items()
    }
    modules_json = json.dumps(modules_dict, indent=2)
    logging.debug("All imported modules and their locations:\n%s", modules_json)


def _convert_rust_locations_to_python_locations(
    imported_modules: dict[str, list[RustLocation]],
) -> dict[str, list[Location]]:
    converted_modules: dict[str, list[Location]] = {}
    for module, locations in imported_modules.items():
        converted_modules[module] = [Location.from_rust_location_object(loc) for loc in locations]
    return converted_modules



from __future__ import annotations

from pathlib import Path

import pytest

from deptry.python_file_finder import get_all_python_files_in
from tests.utils import create_files, run_within_dir


def test_simple(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
            Path("other_dir/subdir/file1.py"),
            Path("other_dir/subdir/file2.py"),
        ])

        files = get_all_python_files_in(
            (Path(),),
            exclude=(".venv",),
            extend_exclude=("other_dir",),
            using_default_exclude=False,
        )

        assert sorted(files) == [
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
        ]


def test_only_matches_start(tmp_path: Path) -> None:
    """
    Test that adding 'subdir' as exclude argument does not also exclude dir/subdir.
    """
    with run_within_dir(tmp_path):
        create_files([
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
            Path("subdir/file1.py"),
            Path("subdir/file2.py"),
        ])

        files = get_all_python_files_in(
            (Path(),), exclude=("foo",), extend_exclude=("subdir",), using_default_exclude=False
        )

        assert sorted(files) == [
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
        ]


@pytest.mark.parametrize(
    ("ignore_notebooks", "expected"),
    [
        (
            False,
            [Path("dir/subdir/file1.ipynb")],
        ),
        (
            True,
            [],
        ),
    ],
)
def test_matches_ipynb(ignore_notebooks: bool, expected: list[Path], tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([Path("dir/subdir/file1.ipynb")])

        files = get_all_python_files_in(
            (Path(),), exclude=(), extend_exclude=(), using_default_exclude=False, ignore_notebooks=ignore_notebooks
        )

        assert sorted(files) == expected


@pytest.mark.parametrize(
    ("exclude", "expected"),
    [
        (
            (".*file1",),
            [
                Path(".cache/file2.py"),
                Path("dir/subdir/file2.py"),
                Path("dir/subdir/file3.py"),
                Path("other_dir/subdir/file2.py"),
            ],
        ),
        (
            (".cache|other.*subdir",),
            [
                Path("dir/subdir/file1.py"),
                Path("dir/subdir/file2.py"),
                Path("dir/subdir/file3.py"),
            ],
        ),
        (
            (".*/subdir/",),
            [
                Path(".cache/file1.py"),
                Path(".cache/file2.py"),
            ],
        ),
    ],
)
def test_regex_argument(exclude: tuple[str], expected: list[Path], tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([
            Path(".cache/file1.py"),
            Path(".cache/file2.py"),
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
            Path("other_dir/subdir/file1.py"),
            Path("other_dir/subdir/file2.py"),
        ])

        files = get_all_python_files_in((Path(),), exclude=exclude, extend_exclude=(), using_default_exclude=False)

        assert sorted(files) == expected


@pytest.mark.parametrize(
    ("exclude", "expected"),
    [
        (
            (".*file1",),
            [
                Path("dir/subdir/file2.py"),
                Path("dir/subdir/file3.py"),
                Path("other_dir/subdir/file2.py"),
            ],
        ),
        (
            (".*file1|.*file2",),
            [
                Path("dir/subdir/file3.py"),
            ],
        ),
        (
            (".*/subdir/",),
            [],
        ),
    ],
)
def test_multiple_source_directories(exclude: tuple[str], expected: list[Path], tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
            Path("other_dir/subdir/file1.py"),
            Path("other_dir/subdir/file2.py"),
            Path("another_dir/subdir/file1.py"),
        ])

        files = get_all_python_files_in(
            (Path("dir"), Path("other_dir")), exclude=exclude, extend_exclude=(), using_default_exclude=False
        )

        assert sorted(files) == expected


def test_duplicates_are_removed(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([Path("dir/subdir/file1.py")])

        files = get_all_python_files_in((Path(), Path()), exclude=(), extend_exclude=(), using_default_exclude=False)

        assert sorted(files) == [Path("dir/subdir/file1.py")]



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    from deptry.dependency import Dependency
    from deptry.imports.location import Location
    from deptry.module import Module, ModuleLocations


@dataclass
class ViolationsFinder(ABC):
    """Base class for all issues finders.

    This abstract class provides a common interface for classes that find issues related to project dependencies.
    Subclasses must implement the 'find' method, which returns a list of Violation objects representing the issues found.

    Attributes:
        imported_modules_with_locations: A list of ModuleLocations objects representing the
            modules imported by the project and their locations.
        dependencies: A list of Dependency objects representing the project's dependencies.
        ignored_modules: A tuple of module names to ignore when scanning for issues. Defaults to an
            empty tuple.
        standard_library_modules: A set of modules that are part of the standard library
    """

    violation: ClassVar[type[Violation]]
    imported_modules_with_locations: list[ModuleLocations]
    dependencies: list[Dependency]
    standard_library_modules: frozenset[str]
    ignored_modules: tuple[str, ...] = ()

    @abstractmethod
    def find(self) -> list[Violation]:
        """Find issues about dependencies."""
        raise NotImplementedError()


@dataclass
class Violation(ABC):
    """
    An abstract base class representing a violation found in the project's dependencies.

    Attributes:
        error_code: A class variable representing the error code associated with the violation. e.g.
            `DEP001`.
        error_template: A class variable representing a string template used to generate an error
            message for the violation.
        issue: An attribute representing the module or dependency where the violation
            occurred.
        location: An attribute representing the location in the code where the violation occurred.
    """

    error_code: ClassVar[str] = ""
    error_template: ClassVar[str] = ""
    issue: Dependency | Module
    location: Location

    @abstractmethod
    def get_error_message(self) -> str:
        raise NotImplementedError()



from __future__ import annotations

import uuid
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from tests.functional.utils import Project
from tests.utils import get_issues_report

if TYPE_CHECKING:
    from tests.utils import PoetryVenvFactory


@pytest.mark.xdist_group(name=Project.POETRY)
def test_cli_with_poetry(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory(Project.POETRY) as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'black' imported but declared as a dev dependency",
                },
                "module": "black",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 4,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'mypy' imported but declared as a dev dependency",
                },
                "module": "mypy",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 6,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'pytest' imported but declared as a dev dependency",
                },
                "module": "pytest",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 7,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'pytest_cov' imported but declared as a dev dependency",
                },
                "module": "pytest_cov",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 8,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP001",
                    "message": "'white' imported but missing from the dependency definitions",
                },
                "module": "white",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 9,
                    "column": 8,
                },
            },
        ]



from __future__ import annotations

from pathlib import Path

from deptry.dependency import Dependency
from deptry.imports.location import Location
from deptry.module import ModuleBuilder, ModuleLocations
from deptry.violations import DEP002UnusedDependenciesFinder, DEP002UnusedDependencyViolation


def test_simple() -> None:
    dependency_toml = Dependency("toml", Path("pyproject.toml"))
    dependencies = [Dependency("click", Path("pyproject.toml")), dependency_toml]
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("click", {"foo"}, frozenset(), dependencies).build(), [Location(Path("foo.py"), 1, 2)]
        )
    ]

    assert DEP002UnusedDependenciesFinder(modules_locations, dependencies, frozenset()).find() == [
        DEP002UnusedDependencyViolation(dependency_toml, Location(Path("pyproject.toml")))
    ]


def test_simple_with_ignore() -> None:
    dependencies = [Dependency("click", Path("pyproject.toml")), Dependency("toml", Path("pyproject.toml"))]
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("toml", {"foo"}, frozenset(), dependencies).build(), [Location(Path("foo.py"), 1, 2)]
        )
    ]

    assert (
        DEP002UnusedDependenciesFinder(modules_locations, dependencies, frozenset(), ignored_modules=("click",)).find()
        == []
    )


def test_top_level() -> None:
    """
    Test if top-level information is read, and correctly used to not mark a dependency as unused.
    blackd is in the top-level of black, so black should not be marked as an unused dependency.
    """
    dependencies = [Dependency("black", Path("pyproject.toml"), module_names=("black", "blackd"))]
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("blackd", {"foo"}, frozenset(), dependencies).build(), [Location(Path("foo.py"), 1, 2)]
        )
    ]

    deps = DEP002UnusedDependenciesFinder(modules_locations, dependencies, frozenset()).find()

    assert deps == []


def test_without_top_level() -> None:
    """
    Test if packages without top-level information are correctly maked as unused
    """
    dependencies = [Dependency("isort", Path("pyproject.toml"))]
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("isort", {"foo"}, frozenset(), dependencies).build(), [Location(Path("foo.py"), 1, 2)]
        )
    ]

    assert DEP002UnusedDependenciesFinder(modules_locations, dependencies, frozenset()).find() == []



from __future__ import annotations

from pathlib import Path

from deptry.dependency import Dependency
from deptry.imports.location import Location
from deptry.module import ModuleBuilder, ModuleLocations
from deptry.violations import DEP001MissingDependenciesFinder, DEP001MissingDependencyViolation


def test_simple() -> None:
    dependencies: list[Dependency] = []

    module_foobar_locations = [Location(Path("foo.py"), 1, 2), Location(Path("bar.py"), 3, 4)]
    module_foobar = ModuleBuilder("foobar", {"foo"}, frozenset(), dependencies).build()

    modules_locations = [ModuleLocations(module_foobar, module_foobar_locations)]

    assert DEP001MissingDependenciesFinder(modules_locations, dependencies, frozenset()).find() == [
        DEP001MissingDependencyViolation(module_foobar, location) for location in module_foobar_locations
    ]


def test_local_module() -> None:
    dependencies: list[Dependency] = []
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("foobar", {"foo", "foobar"}, frozenset(), dependencies).build(),
            [Location(Path("foo.py"), 1, 2)],
        )
    ]

    assert DEP001MissingDependenciesFinder(modules_locations, dependencies, frozenset()).find() == []


def test_simple_with_ignore() -> None:
    dependencies: list[Dependency] = []
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("foobar", {"foo", "bar"}, frozenset(), dependencies).build(), [Location(Path("foo.py"), 1, 2)]
        )
    ]

    assert (
        DEP001MissingDependenciesFinder(
            modules_locations, dependencies, frozenset(), ignored_modules=("foobar",)
        ).find()
        == []
    )


def test_simple_with_standard_library() -> None:
    dependencies: list[Dependency] = []
    standard_library_modules = frozenset(["foobar"])
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("foobar", set(), standard_library_modules, dependencies).build(),
            [Location(Path("foo.py"), 1, 2)],
        )
    ]

    assert DEP001MissingDependenciesFinder(modules_locations, dependencies, frozenset()).find() == []


def test_no_error() -> None:
    """
    This should run without an error, even though `foo` is not installed.
    """

    dependencies = [Dependency("foo", Path("pyproject.toml"))]
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("foo", {"bar"}, frozenset(), dependencies).build(),
            [Location(Path("foo.py"), 1, 2)],
        )
    ]

    assert DEP001MissingDependenciesFinder(modules_locations, dependencies, frozenset()).find() == []



from __future__ import annotations

from enum import Enum

DEPTRY_WHEEL_DIRECTORY = "build/functional_tests/deptry"


class Project(str, Enum):
    EXAMPLE = "example_project"
    PEP_621 = "pep_621_project"
    FUTURE_DEPRECATED_OBSOLETE_ARGUMENT = "project_with_future_deprecated_obsolete_argument"
    GITIGNORE = "project_with_gitignore"
    MULTIPLE_SOURCE_DIRECTORIES = "project_with_multiple_source_directories"
    NAMESPACE = "project_using_namespace"
    PDM = "project_with_pdm"
    POETRY = "project_with_poetry"
    PYPROJECT_DIFFERENT_DIRECTORY = "project_with_pyproject_different_directory"
    REQUIREMENTS_TXT = "project_with_requirements_txt"
    REQUIREMENTS_IN = "project_with_requirements_in"
    SRC_DIRECTORY = "project_with_src_directory"

    def __str__(self) -> str:
        return self.value



from __future__ import annotations

import json
import logging
import re
import uuid
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from deptry.imports.extract import get_imported_modules_from_list_of_files
from deptry.imports.location import Location
from tests.utils import run_within_dir

if TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


def test_import_parser_py() -> None:
    some_imports_path = Path("tests/data/some_imports.py")

    assert get_imported_modules_from_list_of_files([some_imports_path]) == {
        "barfoo": [Location(some_imports_path, 21, 8)],
        "baz": [Location(some_imports_path, 17, 5)],
        "click": [Location(some_imports_path, 31, 12)],
        "foobar": [Location(some_imports_path, 19, 12)],
        "httpx": [Location(some_imports_path, 15, 12)],
        "module_in_class": [Location(some_imports_path, 42, 16)],
        "module_in_func": [Location(some_imports_path, 37, 12)],
        "not_click": [Location(some_imports_path, 33, 12)],
        "numpy": [
            Location(some_imports_path, 6, 8),
            Location(some_imports_path, 8, 1),
        ],
        "os": [Location(some_imports_path, 2, 1)],
        "pandas": [Location(some_imports_path, 7, 8)],
        "pathlib": [Location(some_imports_path, 3, 1)],
        "randomizer": [Location(some_imports_path, 22, 1)],
        "typing": [
            Location(some_imports_path, 1, 8),
            Location(some_imports_path, 4, 1),
        ],
    }


def test_import_parser_ipynb() -> None:
    notebook_path = Path("tests/data/example_project/src/notebook.ipynb")

    assert get_imported_modules_from_list_of_files([notebook_path]) == {
        "click": [Location(notebook_path, 4, 8)],
        "toml": [Location(notebook_path, 6, 8)],
        "urllib3": [Location(notebook_path, 5, 1)],
    }


@pytest.mark.parametrize(
    ("file_content", "encoding"),
    [
        (
            "# -*- encoding: utf-8 -*-\nimport foo\nprint('嘉大')",
            "utf-8",
        ),
        (
            "# -*- encoding: iso-8859-15 -*-\nimport foo\nprint('Æ	Ç')",
            "iso-8859-15",
        ),
        (
            "# -*- encoding: utf-16 -*-\nimport foo\nprint('嘉大')",
            "utf-16",
        ),
        (
            "\nimport foo\nmy_string = '🐺'",
            "utf-16",
        ),
    ],
)
def test_import_parser_file_encodings(file_content: str, encoding: str | None, tmp_path: Path) -> None:
    random_file = Path(f"file_{uuid.uuid4()}.py")

    with run_within_dir(tmp_path):
        with random_file.open("w", encoding=encoding) as f:
            f.write(file_content)

        assert get_imported_modules_from_list_of_files([random_file]) == {"foo": [Location(random_file, 2, 8)]}


@pytest.mark.parametrize(
    ("code_cell_content", "encoding"),
    [
        (
            ["import foo", "print('嘉大')"],
            "utf-8",
        ),
        (
            ["import foo", "print('Æ	Ç')"],
            "iso-8859-15",
        ),
        (
            ["import foo", "print('嘉大')"],
            "utf-16",
        ),
        (
            ["import foo", "my_string = '🐺'"],
            None,
        ),
    ],
)
def test_import_parser_file_encodings_ipynb(code_cell_content: list[str], encoding: str | None, tmp_path: Path) -> None:
    random_file = Path(f"file_{uuid.uuid4()}.ipynb")

    with run_within_dir(tmp_path):
        with random_file.open("w", encoding=encoding) as f:
            file_content = {
                "cells": [
                    {
                        "cell_type": "code",
                        "metadata": {},
                        "source": code_cell_content,
                    }
                ]
            }
            f.write(json.dumps(file_content))

        assert get_imported_modules_from_list_of_files([random_file]) == {"foo": [Location(random_file, 1, 8)]}


def test_import_parser_errors(tmp_path: Path, caplog: LogCaptureFixture) -> None:
    file_ok = Path("file_ok.py")
    file_with_bad_encoding = Path("file_with_bad_encoding.py")
    file_with_syntax_error = Path("file_with_syntax_error.py")

    with run_within_dir(tmp_path):
        with file_ok.open("w") as f:
            f.write("import black")

        with file_with_bad_encoding.open("w", encoding="ISO-8859-1") as f:
            f.write("# -*- coding: utf-8 -*-\nprint('ÆØÅ')")

        with file_with_syntax_error.open("w") as f:
            f.write("invalid_syntax:::")

        with caplog.at_level(logging.WARNING):
            assert get_imported_modules_from_list_of_files([
                file_ok,
                file_with_bad_encoding,
                file_with_syntax_error,
            ]) == {"black": [Location(file=Path("file_ok.py"), line=1, column=8)]}

        assert re.search(
            r"WARNING  .*:shared.rs:\d+ Warning: Skipping processing of file_with_bad_encoding.py because of the following error: \"OSError: Failed to decode file content with the detected encoding.\".",
            caplog.text,
        )
        assert re.search(
            r"WARNING  .*:shared.rs:\d+ Warning: Skipping processing of file_with_syntax_error.py because of the following error: \"SyntaxError: Expected an expression at byte range 15..16\".",
            caplog.text,
        )


def test_import_parser_for_ipynb_errors(tmp_path: Path, caplog: LogCaptureFixture) -> None:
    notebook_ok = Path("notebook_ok.ipynb")
    notebook_with_syntax_error = Path("notebook_with_syntax_error.ipynb")

    with run_within_dir(tmp_path):
        # Create a well-formed notebook
        with notebook_ok.open("w") as f:
            json.dump(
                {
                    "cells": [{"cell_type": "code", "source": ["import numpy\n"]}],
                    "metadata": {},
                    "nbformat": 4,
                    "nbformat_minor": 2,
                },
                f,
            )

        # Create a notebook with invalid Python syntax in a code cell
        with notebook_with_syntax_error.open("w") as f:
            json.dump(
                {
                    "cells": [{"cell_type": "code", "source": ["import n invalid_syntax:::\n"]}],
                    "metadata": {},
                    "nbformat": 4,
                    "nbformat_minor": 2,
                },
                f,
            )

        # Execute function and assert the result for well-formed notebook
        with caplog.at_level(logging.WARNING):
            assert get_imported_modules_from_list_of_files([
                notebook_ok,
                notebook_with_syntax_error,
            ]) == {"numpy": [Location(file=Path("notebook_ok.ipynb"), line=1, column=8)]}

        assert re.search(
            r"WARNING  .*:shared.rs:\d+ Warning: Skipping processing of notebook_with_syntax_error.ipynb because of the following error: \"SyntaxError: Expected ',', found name at byte range 9..23\".",
            caplog.text,
        )


def test_python_3_12_f_string_syntax(tmp_path: Path) -> None:
    file_path = Path("file1.py")

    with run_within_dir(tmp_path):
        with file_path.open("w") as f:
            f.write('import foo\nprint(f"abc{"def"}")')

        assert get_imported_modules_from_list_of_files([file_path]) == {"foo": [Location(file_path, 1, 8)]}



from __future__ import annotations

import re
from typing import TYPE_CHECKING
from unittest import mock

import click
import pytest

from deptry.cli import CommaSeparatedMappingParamType, CommaSeparatedTupleParamType

if TYPE_CHECKING:
    from collections.abc import MutableMapping, Sequence
    from re import Pattern


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        pytest.param(
            "",
            ("",),
            id="from empty str",
        ),
        pytest.param(
            "foo",
            ("foo",),
            id="from single str",
        ),
        pytest.param(
            "foo,bar",
            ("foo", "bar"),
            id="from multiple str",
        ),
        pytest.param(
            [],
            (),
            id="from empty list",
        ),
        pytest.param(
            ["foo"],
            ("foo",),
            id="from single item list",
        ),
        pytest.param(
            ["foo", "bar"],
            ("foo", "bar"),
            id="from multiple item list",
        ),
        pytest.param(
            (),
            (),
            id="from empty tuple",
        ),
        pytest.param(
            ("foo",),
            ("foo",),
            id="from single item tuple",
        ),
        pytest.param(
            ("foo", "bar"),
            ("foo", "bar"),
            id="from multiple item tuple",
        ),
    ],
)
def test_comma_separated_tuple_param_type_convert(
    value: str | list[str] | tuple[str, ...],
    expected: tuple[str, ...],
) -> None:
    """Tests all valid conversion paths."""
    param = mock.Mock(spec=click.Parameter)
    ctx = mock.Mock(spec=click.Context)

    actual = CommaSeparatedTupleParamType().convert(value=value, param=param, ctx=ctx)

    assert actual == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        pytest.param(
            "foo=bar",
            {"foo": ("bar",)},
            id="from str, single key, single value",
        ),
        pytest.param(
            "foo=bar|baz",
            {"foo": ("bar", "baz")},
            id="from str, single key, multi value",
        ),
        pytest.param(
            "foo=bar,fox=fuz",
            {"foo": ("bar",), "fox": ("fuz",)},
            id="from str, multi key, single value",
        ),
        pytest.param(
            "foo=bar,foo=fuz",
            {"foo": ("bar", "fuz")},
            id="from str, redefined key, single value",
        ),
        pytest.param(
            {"foo": "bar"},
            {"foo": ("bar",)},
            id="from str-to-str map",
        ),
        pytest.param(
            "foo|bar=baz",
            {"foo|bar": ("baz",)},
            id="from str, key with multiple-value separator token",
        ),
        pytest.param(
            "foo=bar=baz",
            {"foo": ("bar=baz",)},
            id="from str, single value with key-value seperator token",
        ),
        pytest.param(
            "foo=",
            {"foo": ("",)},
            id="from str, empty value",
        ),
        pytest.param(
            {"foo": ("bar",)},
            {"foo": ("bar",)},
            id="from str-to-tuple-of-str map",
        ),
        pytest.param(
            {"foo": ("bar",), "bar": "baz"},
            {"foo": ("bar",), "bar": ("baz",)},
            id="from str-to-mixed map",
        ),
    ],
)
def test_comma_separated_mapping_param_type_convert(
    value: str | MutableMapping[str, Sequence[str] | str],
    expected: MutableMapping[str, tuple[str, ...]],
) -> None:
    """Tests all valid conversion paths."""
    param = mock.Mock(spec=click.Parameter)
    ctx = mock.Mock(spec=click.Context)

    actual = CommaSeparatedMappingParamType().convert(value=value, param=param, ctx=ctx)

    assert actual == expected


@pytest.mark.parametrize(
    ("value", "err_type", "err_msg_matcher"),
    [
        pytest.param(
            "",
            ValueError,
            re.compile(r"equal sign"),
            id="from empty str",
        ),
        pytest.param(
            "foo,bar=baz",
            ValueError,
            re.compile(r"equal sign"),
            id="from invalid first item",
        ),
    ],
)
def test_comma_separated_mapping_param_type_convert_err(
    value: str | MutableMapping[str, Sequence[str] | str],
    err_type: type[Exception],
    err_msg_matcher: Pattern[str] | str,
) -> None:
    """Tests some invalid conversion paths."""
    param = mock.Mock(spec=click.Parameter)
    ctx = mock.Mock(spec=click.Context)
    param_type = CommaSeparatedMappingParamType()

    with pytest.raises(err_type, match=err_msg_matcher):
        param_type.convert(value=value, param=param, ctx=ctx)



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Mapping

from deptry.dependency_getter.pdm import PDMDependencyGetter
from deptry.dependency_getter.pep_621 import PEP621DependencyGetter
from deptry.dependency_getter.poetry import PoetryDependencyGetter
from deptry.dependency_getter.requirements_files import RequirementsTxtDependencyGetter
from deptry.exceptions import DependencySpecificationNotFoundError
from deptry.utils import load_pyproject_toml

if TYPE_CHECKING:
    from typing import Any

    from deptry.dependency_getter.base import DependencyGetter


@dataclass
class DependencyGetterBuilder:
    """
    Class to detect how dependencies are specified:
    - Either find a pyproject.toml with a [poetry.tool.dependencies] section
    - Otherwise, find a pyproject.toml with a [tool.pdm] section
    - Otherwise, find a pyproject.toml with a [project] section
    - Otherwise, find a requirements.txt.
    """

    config: Path
    package_module_name_map: Mapping[str, tuple[str, ...]] = field(default_factory=dict)
    pep621_dev_dependency_groups: tuple[str, ...] = ()
    requirements_files: tuple[str, ...] = ()
    using_default_requirements_files: bool = True
    requirements_files_dev: tuple[str, ...] = ()

    def build(self) -> DependencyGetter:
        pyproject_toml_found = self._project_contains_pyproject_toml()

        if pyproject_toml_found:
            pyproject_toml = load_pyproject_toml(self.config)

            if self._project_uses_poetry(pyproject_toml):
                return PoetryDependencyGetter(self.config, self.package_module_name_map)

            if self._project_uses_pdm(pyproject_toml):
                return PDMDependencyGetter(self.config, self.package_module_name_map, self.pep621_dev_dependency_groups)

            if self._project_uses_pep_621(pyproject_toml):
                return PEP621DependencyGetter(
                    self.config, self.package_module_name_map, self.pep621_dev_dependency_groups
                )

        check, requirements_files = self._project_uses_requirements_files()
        if check:
            return RequirementsTxtDependencyGetter(
                self.config, self.package_module_name_map, requirements_files, self.requirements_files_dev
            )

        raise DependencySpecificationNotFoundError(self.requirements_files)

    def _project_contains_pyproject_toml(self) -> bool:
        if self.config.exists():
            logging.debug("pyproject.toml found!")
            return True
        else:
            logging.debug("No pyproject.toml found.")
            return False

    @staticmethod
    def _project_uses_poetry(pyproject_toml: dict[str, Any]) -> bool:
        try:
            pyproject_toml["tool"]["poetry"]
            logging.debug(
                "pyproject.toml contains a [tool.poetry] section, so Poetry is used to specify the"
                " project's dependencies."
            )
        except KeyError:
            logging.debug(
                "pyproject.toml does not contain a [tool.poetry] section, so Poetry is not used to specify"
                " the project's dependencies."
            )
            return False
        else:
            return True

    @staticmethod
    def _project_uses_pdm(pyproject_toml: dict[str, Any]) -> bool:
        try:
            pyproject_toml["tool"]["pdm"]["dev-dependencies"]
            logging.debug(
                "pyproject.toml contains a [tool.pdm.dev-dependencies] section, so PDM is used to specify the project's"
                " dependencies."
            )
        except KeyError:
            logging.debug(
                "pyproject.toml does not contain a [tool.pdm.dev-dependencies] section, so PDM is not used to specify"
                " the project's dependencies."
            )
            return False
        else:
            return True

    @staticmethod
    def _project_uses_pep_621(pyproject_toml: dict[str, Any]) -> bool:
        if pyproject_toml.get("project"):
            logging.debug(
                "pyproject.toml contains a [project] section, so PEP 621 is used to specify the project's dependencies."
            )
            return True

        logging.debug(
            "pyproject.toml does not contain a [project] section, so PEP 621 is not used to specify the project's"
            " dependencies."
        )
        return False

    def _project_uses_requirements_files(self) -> tuple[bool, tuple[str, ...]]:
        """
        Tools like `pip-tools` and `uv` work with a setup in which a `requirements.in` is compiled into a `requirements.txt`, which then
        contains pinned versions for all transitive dependencies. If the user did not explicitly specify the argument `requirements-files`,
        but there is a `requirements.in` present, it is highly likely that the user wants to use the `requirements.in` file so we set
        `requirements-files` to that instead.
        """
        if self.using_default_requirements_files and Path("requirements.in").is_file():
            logging.info(
                "Detected a 'requirements.in' file in the project and no 'requirements-files' were explicitly specified. "
                "Automatically using 'requirements.in' as the source for the project's dependencies. To specify a different source for "
                "the project's dependencies, use the '--requirements-files' option."
            )
            return True, ("requirements.in",)

        check = any(Path(requirements_files).is_file() for requirements_files in self.requirements_files)
        if check:
            logging.debug(
                "Dependency specification found in '%s'. Will use this to determine the project's dependencies.\n",
                ", ".join(self.requirements_files),
            )
            return True, self.requirements_files
        return False, ()



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

from deptry.dependency_getter.pep_621 import PEP621DependencyGetter
from tests.utils import run_within_dir

if TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


def test_dependency_getter(tmp_path: Path) -> None:
    fake_pyproject_toml = """[project]
# PEP 621 project metadata
# See https://www.python.org/dev/peps/pep-0621/
dependencies = [
    "qux",
    "bar>=20.9",
    "optional-foo[option]>=0.12.11",
    "conditional-bar>=1.1.0; python_version < 3.11",
    "fox-python",  # top level module is called "fox"
]

[project.optional-dependencies]
group1 = [
    "foobar",
    "barfoo",
]
group2 = [
    "dep",
]
"""

    with run_within_dir(tmp_path):
        with Path("pyproject.toml").open("w") as f:
            f.write(fake_pyproject_toml)

        getter = PEP621DependencyGetter(
            config=Path("pyproject.toml"),
            package_module_name_map={"fox-python": ("fox",)},
        )
        dependencies = getter.get().dependencies

        assert len(dependencies) == 8

        assert dependencies[0].name == "qux"
        assert "qux" in dependencies[0].top_levels

        assert dependencies[1].name == "bar"
        assert "bar" in dependencies[1].top_levels

        assert dependencies[2].name == "optional-foo"
        assert "optional_foo" in dependencies[2].top_levels

        assert dependencies[3].name == "conditional-bar"
        assert "conditional_bar" in dependencies[3].top_levels

        assert dependencies[4].name == "fox-python"
        assert "fox" in dependencies[4].top_levels

        assert dependencies[5].name == "foobar"
        assert "foobar" in dependencies[5].top_levels

        assert dependencies[6].name == "barfoo"
        assert "barfoo" in dependencies[6].top_levels

        assert dependencies[7].name == "dep"
        assert "dep" in dependencies[7].top_levels


def test_dependency_getter_with_dev_dependencies(tmp_path: Path) -> None:
    fake_pyproject_toml = """[project]
# PEP 621 project metadata
# See https://www.python.org/dev/peps/pep-0621/
dependencies = [
"qux",
]

[project.optional-dependencies]
group1 = [
    "foobar",
]
group2 = [
    "barfoo",
]
"""

    with run_within_dir(tmp_path):
        with Path("pyproject.toml").open("w") as f:
            f.write(fake_pyproject_toml)

        getter = PEP621DependencyGetter(config=Path("pyproject.toml"), pep621_dev_dependency_groups=("group2",))
        dependencies = getter.get().dependencies
        dev_dependencies = getter.get().dev_dependencies

        assert len(dependencies) == 2

        assert dependencies[0].name == "qux"
        assert "qux" in dependencies[0].top_levels

        assert dependencies[1].name == "foobar"
        assert "foobar" in dependencies[1].top_levels

        assert len(dev_dependencies) == 1
        assert dev_dependencies[0].name == "barfoo"
        assert "barfoo" in dev_dependencies[0].top_levels


def test_dependency_getter_with_incorrect_dev_group(tmp_path: Path, caplog: LogCaptureFixture) -> None:
    fake_pyproject_toml = """[project]
# PEP 621 project metadata
# See https://www.python.org/dev/peps/pep-0621/
dependencies = [
"qux",
]

[project.optional-dependencies]
group1 = [
    "foobar",
]
group2 = [
    "barfoo",
]
"""

    with run_within_dir(tmp_path), caplog.at_level(logging.INFO):
        with Path("pyproject.toml").open("w") as f:
            f.write(fake_pyproject_toml)

        getter = PEP621DependencyGetter(config=Path("pyproject.toml"), pep621_dev_dependency_groups=("group3",))
        dependencies = getter.get().dependencies

        assert (
            "Trying to extract the dependencies from the optional dependency groups ['group3'] as development dependencies, but the following groups were not found: ['group3']"
            in caplog.text
        )

        assert len(dependencies) == 3

        assert dependencies[0].name == "qux"
        assert "qux" in dependencies[0].top_levels

        assert dependencies[1].name == "foobar"
        assert "foobar" in dependencies[1].top_levels

        assert dependencies[2].name == "barfoo"
        assert "barfoo" in dependencies[2].top_levels


def test_dependency_getter_empty_dependencies(tmp_path: Path) -> None:
    fake_pyproject_toml = """[project]
# PEP 621 project metadata
# See https://www.python.org/dev/peps/pep-0621/
"""

    with run_within_dir(tmp_path):
        with Path("pyproject.toml").open("w") as f:
            f.write(fake_pyproject_toml)

        getter = PEP621DependencyGetter(config=Path("pyproject.toml"))
        dependencies_extract = getter.get()

        assert len(dependencies_extract.dependencies) == 0
        assert len(dependencies_extract.dev_dependencies) == 0



from __future__ import annotations

from typing import TYPE_CHECKING

from click import UsageError

if TYPE_CHECKING:
    from pathlib import Path


class DependencySpecificationNotFoundError(FileNotFoundError):
    def __init__(self, requirements_files: tuple[str, ...]) -> None:
        super().__init__(
            "No file called 'pyproject.toml' with a [tool.poetry.dependencies], [tool.pdm] or [project] section or"
            f" file(s) called '{', '.join(requirements_files)}' found. Exiting."
        )


class PyprojectFileNotFoundError(FileNotFoundError):
    def __init__(self, directory: Path) -> None:
        super().__init__(f"No file `pyproject.toml` found in directory {directory}")


class UnsupportedPythonVersionError(ValueError):
    def __init__(self, version: tuple[int, int]) -> None:
        super().__init__(
            f"Python version {version[0]}.{version[1]} is not supported. Only versions >= 3.8 are supported."
        )


class InvalidPyprojectTOMLOptionsError(UsageError):
    def __init__(self, invalid_options: list[str]) -> None:
        super().__init__(
            f"'[tool.deptry]' section in 'pyproject.toml' contains invalid configuration options: {invalid_options}."
        )



from __future__ import annotations

from pathlib import Path

import pytest

from deptry.dependency_getter.requirements_files import RequirementsTxtDependencyGetter
from tests.utils import run_within_dir


def test_parse_requirements_files(tmp_path: Path) -> None:
    fake_requirements_files = """click==8.1.3 #123asd
colorama==0.4.5
importlib-metadata==4.2.0 ; python_version >= "3.7" and python_version < "3.8"
isort==5.10.1, <6.0
toml==0.10.2
typing-extensions
zipp==3.8.1
foobar[foo, bar]
SomeProject ~= 1.4.2
SomeProject2 == 5.4 ; python_version < '3.8'
SomeProject3 ; sys_platform == 'win32'
requests [security] >= 2.8.1, == 2.8.* ; python_version < "2.7"
# This is a comment, to show how #-prefixed lines are ignored.
#this is a comment without a space after the `#`, which should also be ignored.
pytest
pytest-cov
beautifulsoup4
docopt == 0.6.1
requests [security] >= 2.8.1, == 2.8.* ; python_version < "2.7"
fox-python
httpx==0.25.2
"""
    with run_within_dir(tmp_path):
        with Path("requirements.txt").open("w") as f:
            f.write(fake_requirements_files)

        getter = RequirementsTxtDependencyGetter(
            config=Path("pyproject.toml"),
            package_module_name_map={"fox-python": ("fox",)},
        )
        dependencies_extract = getter.get()
        dependencies = dependencies_extract.dependencies

        assert len(dependencies) == 19
        assert len(dependencies_extract.dev_dependencies) == 0

        assert dependencies[1].name == "colorama"
        assert "colorama" in dependencies[1].top_levels

        assert dependencies[2].name == "importlib-metadata"
        assert "importlib_metadata" in dependencies[2].top_levels

        assert dependencies[11].name == "requests"
        assert "requests" in dependencies[11].top_levels

        assert dependencies[17].name == "fox-python"
        assert "fox" in dependencies[17].top_levels


def test_parse_requirements_files_urls(tmp_path: Path) -> None:
    fake_requirements_files = """urllib3 @ https://github.com/urllib3/urllib3/archive/refs/tags/1.26.8.zip
https://github.com/urllib3/urllib3/archive/refs/tags/1.26.8.zip
git+https://github.com/baz/foo-bar.git@asd#egg=foo-bar
git+https://github.com/baz/foo-bar.git@asd
git+https://github.com/abc123/bar-foo@xyz789#egg=bar-fooo"""

    with run_within_dir(tmp_path):
        with Path("requirements.txt").open("w") as f:
            f.write(fake_requirements_files)

        dependencies_extract = RequirementsTxtDependencyGetter(Path("pyproject.toml")).get()
        dependencies = dependencies_extract.dependencies

        assert len(dependencies) == 5
        assert len(dependencies_extract.dev_dependencies) == 0

        assert dependencies[0].name == "urllib3"
        assert dependencies[1].name == "urllib3"
        assert dependencies[2].name == "foo-bar"
        assert dependencies[3].name == "foo-bar"
        assert dependencies[4].name == "bar-fooo"


def test_single(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        with Path("req.txt").open("w") as f:
            f.write("click==8.1.3 #123asd\ncolorama==0.4.5")

        dependencies_extract = RequirementsTxtDependencyGetter(
            Path("pyproject.toml"), requirements_files=("req.txt",)
        ).get()
        dependencies = dependencies_extract.dependencies

        assert len(dependencies) == 2
        assert len(dependencies_extract.dev_dependencies) == 0

        assert dependencies[0].name == "click"
        assert dependencies[1].name == "colorama"


def test_multiple(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        with Path("foo.txt").open("w") as f:
            f.write("click==8.1.3 #123asd")

        with Path("bar.txt").open("w") as f:
            f.write("bar")

        dependencies_extract = RequirementsTxtDependencyGetter(
            Path("pyproject.toml"), requirements_files=("foo.txt", "bar.txt")
        ).get()
        dependencies = dependencies_extract.dependencies

        assert len(dependencies) == 2
        assert len(dependencies_extract.dev_dependencies) == 0

        assert dependencies[0].name == "click"
        assert dependencies[1].name == "bar"


def test_dev_single(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        with Path("requirements.txt").open("w") as f:
            f.write("")

        with Path("requirements-dev.txt").open("w") as f:
            f.write("click==8.1.3 #123asd\ncolorama==0.4.5")

        dependencies_extract = RequirementsTxtDependencyGetter(Path("pyproject.toml")).get()
        dev_dependencies = dependencies_extract.dev_dependencies

        assert len(dependencies_extract.dependencies) == 0
        assert len(dev_dependencies) == 2

        assert dev_dependencies[1].name == "colorama"
        assert "colorama" in dev_dependencies[1].top_levels


def test_dev_multiple(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        with Path("requirements.txt").open("w") as f:
            f.write("")

        with Path("requirements-dev.txt").open("w") as f:
            f.write("click==8.1.3 #123asd")

        with Path("dev-requirements.txt").open("w") as f:
            f.write("bar")

        dependencies_extract = RequirementsTxtDependencyGetter(Path("pyproject.toml")).get()
        dev_dependencies = dependencies_extract.dev_dependencies

        assert len(dependencies_extract.dependencies) == 0
        assert len(dev_dependencies) == 2

        assert "click" in [dev_dependencies[0].name, dev_dependencies[1].name]
        assert "bar" in [dev_dependencies[0].name, dev_dependencies[1].name]


def test_dev_multiple_with_arguments(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        with Path("requirements.txt").open("w") as f:
            f.write("")

        with Path("foo.txt").open("w") as f:
            f.write("click==8.1.3 #123asd")

        with Path("bar.txt").open("w") as f:
            f.write("bar")

        dependencies_extract = RequirementsTxtDependencyGetter(
            Path("pyproject.toml"), requirements_files_dev=("foo.txt", "bar.txt")
        ).get()
        dev_dependencies = dependencies_extract.dev_dependencies

        assert len(dependencies_extract.dependencies) == 0
        assert len(dev_dependencies) == 2

        assert dev_dependencies[0].name == "click"
        assert dev_dependencies[1].name == "bar"


@pytest.mark.parametrize(
    ("line", "expected"),
    [
        ("foo", False),
        ("http", False),
        ("https", False),
        ("httpx", False),
        ("git+http", False),
        ("git+https", False),
        ("http://", True),
        ("https://", True),
        ("git+http://", True),
        ("git+https://", True),
        ("file://", True),
        ("file:///", True),
        ("httpx://", True),
    ],
)
def test__line_is_url(line: str, expected: bool) -> None:
    assert RequirementsTxtDependencyGetter._line_is_url(line) is expected



from __future__ import annotations

from pathlib import Path

from deptry.rust import find_python_files


def get_all_python_files_in(
    directories: tuple[Path, ...],
    exclude: tuple[str, ...],
    extend_exclude: tuple[str, ...],
    using_default_exclude: bool,
    ignore_notebooks: bool = False,
) -> list[Path]:
    return [
        Path(f)
        for f in find_python_files(directories, exclude, extend_exclude, using_default_exclude, ignore_notebooks)
    ]



from __future__ import annotations

import itertools
import logging
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.dependency import Dependency
from deptry.dependency_getter.base import DependenciesExtract, DependencyGetter
from deptry.utils import load_pyproject_toml

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


@dataclass
class PEP621DependencyGetter(DependencyGetter):
    pep621_dev_dependency_groups: tuple[str, ...] = ()
    """
    Class to extract dependencies from a pyproject.toml file in which dependencies are specified according to PEP 621. For example:

        [project]
        ...
        dependencies = [
            "httpx",
            "gidgethub[httpx]>4.0.0",
            "django>2.1; os_name != 'nt'",
            "django>2.0; os_name == 'nt'"
        ]

        [project.optional-dependencies]
        test = [
            "pytest < 5.0.0",
            "pytest-cov[all]"
        ]

    Note that by default both dependencies and optional-dependencies are extracted as regular dependencies, since PEP-621 does not specify
    a recommended way to extract development dependencies. However, if a value is passed for the `pep621_dev_dependency_groups`
    argument, all dependencies from groups in that argument are considered to be development dependencies. e.g. in the example above, when
    `pep621_dev_dependency_groups=(test,)`, both `pytest` and `pytest-cov` are returned as development dependencies.
    """

    def get(self) -> DependenciesExtract:
        dependencies = self._get_dependencies()
        optional_dependencies = self._get_optional_dependencies()

        if self.pep621_dev_dependency_groups:
            self._check_for_invalid_group_names(optional_dependencies)
            dev_dependencies, leftover_optional_dependencies = (
                self._split_development_dependencies_from_optional_dependencies(optional_dependencies)
            )
            dependencies = [*dependencies, *leftover_optional_dependencies]
            return DependenciesExtract(dependencies, dev_dependencies)

        dependencies = [*dependencies, *itertools.chain(*optional_dependencies.values())]
        return DependenciesExtract(dependencies, [])

    def _get_dependencies(self) -> list[Dependency]:
        pyproject_data = load_pyproject_toml(self.config)
        dependency_strings: list[str] = pyproject_data["project"].get("dependencies", [])
        return self._extract_pep_508_dependencies(dependency_strings, self.package_module_name_map)

    def _get_optional_dependencies(self) -> dict[str, list[Dependency]]:
        pyproject_data = load_pyproject_toml(self.config)

        return {
            group: self._extract_pep_508_dependencies(dependencies, self.package_module_name_map)
            for group, dependencies in pyproject_data["project"].get("optional-dependencies", {}).items()
        }

    def _check_for_invalid_group_names(self, optional_dependencies: dict[str, list[Dependency]]) -> None:
        missing_groups = set(self.pep621_dev_dependency_groups) - set(optional_dependencies.keys())
        if missing_groups:
            logging.warning(
                "Warning: Trying to extract the dependencies from the optional dependency groups %s as development dependencies, "
                "but the following groups were not found: %s",
                list(self.pep621_dev_dependency_groups),
                list(missing_groups),
            )

    def _split_development_dependencies_from_optional_dependencies(
        self, optional_dependencies: dict[str, list[Dependency]]
    ) -> tuple[list[Dependency], list[Dependency]]:
        """
        Split the optional dependencies into optional dependencies and development dependencies based on the `pep621_dev_dependency_groups`
        parameter. Return a tuple with two values: a list of the development dependencies and a list of the remaining 'true' optional dependencies.
        """
        dev_dependencies = list(
            itertools.chain.from_iterable(
                deps for group, deps in optional_dependencies.items() if group in self.pep621_dev_dependency_groups
            )
        )
        regular_dependencies = list(
            itertools.chain.from_iterable(
                deps for group, deps in optional_dependencies.items() if group not in self.pep621_dev_dependency_groups
            )
        )
        return dev_dependencies, regular_dependencies

    def _extract_pep_508_dependencies(
        self, dependencies: list[str], package_module_name_map: Mapping[str, Sequence[str]]
    ) -> list[Dependency]:
        """
        Given a list of dependency specifications (e.g. "django>2.1; os_name != 'nt'"), convert them to Dependency objects.
        """
        extracted_dependencies = []

        for spec in dependencies:
            # An example of a spec is `"tomli>=1.1.0; python_version < \"3.11\""`
            name = self._find_dependency_name_in(spec)
            if name:
                extracted_dependencies.append(
                    Dependency(
                        name,
                        self.config,
                        module_names=package_module_name_map.get(name),
                    )
                )

        return extracted_dependencies

    @staticmethod
    def _find_dependency_name_in(spec: str) -> str | None:
        match = re.search("[a-zA-Z0-9-_]+", spec)
        if match:
            return match.group(0)
        return None



from __future__ import annotations

import logging
import re
import sys
from pathlib import Path
from unittest import mock

import pytest

from deptry.core import Core
from deptry.dependency import Dependency
from deptry.dependency_getter.base import DependenciesExtract
from deptry.exceptions import UnsupportedPythonVersionError
from deptry.imports.location import Location
from deptry.module import Module
from deptry.stdlibs import STDLIBS_PYTHON
from deptry.violations import (
    DEP001MissingDependencyViolation,
    DEP002UnusedDependencyViolation,
    DEP003TransitiveDependencyViolation,
    DEP004MisplacedDevDependencyViolation,
)
from tests.utils import create_files, run_within_dir


@pytest.mark.parametrize(
    ("known_first_party", "root_suffix", "expected"),
    [
        (
            (),
            "",
            {"module_with_init", "module_without_init", "local_file"},
        ),
        (
            ("module_with_init", "module_without_init"),
            "",
            {"module_with_init", "module_without_init", "local_file"},
        ),
        (
            ("module_without_init",),
            "module_with_init",
            {"foo", "module_without_init", "subdirectory"},
        ),
    ],
)
def test__get_local_modules(
    tmp_path: Path, known_first_party: tuple[str, ...], root_suffix: str, expected: set[str]
) -> None:
    with run_within_dir(tmp_path):
        create_files([
            Path("module_with_init/__init__.py"),
            Path("module_with_init/foo.py"),
            Path("module_with_init/subdirectory/__init__.py"),
            Path("module_with_init/subdirectory/foo.py"),
            Path("module_without_init/bar.py"),
            Path("local_file.py"),
        ])

        assert (
            Core(
                root=(tmp_path / root_suffix,),
                config=Path("pyproject.toml"),
                no_ansi=False,
                per_rule_ignores={},
                ignore=(),
                exclude=(),
                extend_exclude=(),
                using_default_exclude=True,
                ignore_notebooks=False,
                requirements_files=(),
                requirements_files_dev=(),
                known_first_party=known_first_party,
                json_output="",
                package_module_name_map={},
                pep621_dev_dependency_groups=(),
                using_default_requirements_files=True,
            )._get_local_modules()
            == expected
        )


@pytest.mark.skipif(sys.version_info >= (3, 10), reason="mapping is only used for Python < 3.10")
def test__get_stdlib_packages_without_stdlib_module_names() -> None:
    assert Core._get_standard_library_modules() == STDLIBS_PYTHON[f"{sys.version_info[0]}{sys.version_info[1]}"]


@pytest.mark.skipif(sys.version_info < (3, 10), reason="only Python >= 3.10 has sys.stdlib_module_names")
def test__get_stdlib_packages_with_stdlib_module_names() -> None:
    assert Core._get_standard_library_modules() == sys.stdlib_module_names  # type: ignore[attr-defined, unused-ignore]


@pytest.mark.parametrize(
    "version_info",
    [
        (sys.version_info[0], sys.version_info[1] + 1, 0),
        (sys.version_info[0], sys.version_info[1] + 1, 13),
        (sys.version_info[0] + 1, sys.version_info[1], 0),
        (sys.version_info[0], sys.version_info[1] + 1, 0, "beta", 1),
        (sys.version_info[0], sys.version_info[1] + 1, 0, "candidate", 1),
    ],
)
@pytest.mark.skipif(sys.version_info < (3, 10), reason="only Python >= 3.10 has sys.stdlib_module_names")
def test__get_stdlib_packages_with_stdlib_module_names_future_version(version_info: tuple[int | str, ...]) -> None:
    """Test that future versions of Python not yet tested on the CI will also work."""
    with mock.patch("sys.version_info", (sys.version_info[0], sys.version_info[1] + 1, 0)):
        assert Core._get_standard_library_modules() == sys.stdlib_module_names  # type: ignore[attr-defined, unused-ignore]


@pytest.mark.parametrize(
    "version_info",
    [
        (2, 1, 0),
        (2, 7, 0),
        (2, 7, 15),
        (3, 6, 0),
        (3, 6, 7),
        (3, 6, 7, "candidate", 1),
    ],
)
def test__get_stdlib_packages_unsupported(version_info: tuple[int | str, ...]) -> None:
    """It should raise an error when Python version is unsupported."""
    with mock.patch("sys.version_info", version_info), pytest.raises(
        UnsupportedPythonVersionError,
        match=re.escape(
            f"Python version {version_info[0]}.{version_info[1]} is not supported. Only versions >= 3.8 are supported."
        ),
    ):
        Core._get_standard_library_modules()


def test__exit_with_violations() -> None:
    violations = [
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 2)),
        DEP002UnusedDependencyViolation(Dependency("foo", Path("pyproject.toml")), Location(Path("pyproject.toml"))),
        DEP003TransitiveDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 2)),
        DEP004MisplacedDevDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 2)),
    ]

    with pytest.raises(SystemExit) as e:
        Core._exit(violations)

    assert e.type is SystemExit
    assert e.value.code == 1


def test__exit_without_violations() -> None:
    with pytest.raises(SystemExit) as e:
        Core._exit([])

    assert e.type is SystemExit
    assert e.value.code == 0


@pytest.mark.parametrize(
    ("dependencies", "dev_dependencies", "expected_logs"),
    [
        (
            [],
            [],
            [],
        ),
        (
            [
                Dependency("foo", Path("pyproject.toml")),
                Dependency("bar", Path("pyproject.toml")),
            ],
            [
                Dependency("dev", Path("pyproject.toml")),
                Dependency("another-dev", Path("pyproject.toml")),
            ],
            [
                "The project contains the following dependencies:",
                "Dependency 'foo' with top-levels: {'foo'}.",
                "Dependency 'bar' with top-levels: {'bar'}.",
                "",
                "The project contains the following dev dependencies:",
                "Dependency 'dev' with top-levels: {'dev'}.",
                "Dependency 'another-dev' with top-levels: {'another_dev'}.",
                "",
            ],
        ),
    ],
)
def test__log_dependencies(
    dependencies: list[Dependency],
    dev_dependencies: list[Dependency],
    expected_logs: list[str],
    caplog: pytest.LogCaptureFixture,
) -> None:
    with caplog.at_level(logging.DEBUG):
        Core._log_dependencies(DependenciesExtract(dependencies, dev_dependencies))

    assert caplog.messages == expected_logs



from pathlib import Path

from .rust import Location as RustLocation

def get_imports_from_py_files(file_paths: list[str]) -> dict[str, list[RustLocation]]: ...
def get_imports_from_ipynb_files(file_paths: list[str]) -> dict[str, list[RustLocation]]: ...
def find_python_files(
    directories: tuple[Path, ...],
    exclude: tuple[str, ...],
    extend_exclude: tuple[str, ...],
    using_default_exclude: bool,
    ignore_notebooks: bool = False,
) -> list[str]: ...

class Location:
    file: str
    line: int
    column: int
    def __init__(self, file: str, line: int, column: int) -> None: ...



from __future__ import annotations

import json
from pathlib import Path

from deptry.dependency import Dependency
from deptry.imports.location import Location
from deptry.module import Module
from deptry.reporters import JSONReporter
from deptry.violations import (
    DEP001MissingDependencyViolation,
    DEP002UnusedDependencyViolation,
    DEP003TransitiveDependencyViolation,
    DEP004MisplacedDevDependencyViolation,
)
from tests.utils import run_within_dir


def test_simple(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        JSONReporter(
            [
                DEP001MissingDependencyViolation(Module("foo", package="foo-package"), Location(Path("foo.py"), 1, 2)),
                DEP002UnusedDependencyViolation(
                    Dependency("foo", Path("pyproject.toml")), Location(Path("pyproject.toml"))
                ),
                DEP003TransitiveDependencyViolation(
                    Module("foo", package="foo-package"), Location(Path("foo/bar.py"), 1, 2)
                ),
                DEP004MisplacedDevDependencyViolation(
                    Module("foo", package="foo-package"), Location(Path("foo.py"), 1, 2)
                ),
            ],
            "output.json",
        ).report()

        with Path("output.json").open() as f:
            data = json.load(f)

        assert data == [
            {
                "error": {"code": "DEP001", "message": "'foo' imported but missing from the dependency definitions"},
                "module": "foo",
                "location": {
                    "file": str(Path("foo.py")),
                    "line": 1,
                    "column": 2,
                },
            },
            {
                "error": {"code": "DEP002", "message": "'foo' defined as a dependency but not used in the codebase"},
                "module": "foo",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {"code": "DEP003", "message": "'foo' imported but it is a transitive dependency"},
                "module": "foo",
                "location": {
                    "file": str(Path("foo/bar.py")),
                    "line": 1,
                    "column": 2,
                },
            },
            {
                "error": {"code": "DEP004", "message": "'foo' imported but declared as a dev dependency"},
                "module": "foo",
                "location": {
                    "file": str(Path("foo.py")),
                    "line": 1,
                    "column": 2,
                },
            },
        ]



from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from deptry.dependency_getter.builder import DependencyGetterBuilder
from deptry.dependency_getter.pdm import PDMDependencyGetter
from deptry.dependency_getter.pep_621 import PEP621DependencyGetter
from deptry.dependency_getter.poetry import PoetryDependencyGetter
from deptry.dependency_getter.requirements_files import RequirementsTxtDependencyGetter
from deptry.exceptions import DependencySpecificationNotFoundError
from tests.utils import run_within_dir

if TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


def test_poetry(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write('[tool.poetry.dependencies]\nfake = "10"')

        spec = DependencyGetterBuilder(pyproject_toml_path).build()
        assert isinstance(spec, PoetryDependencyGetter)


def test_pdm_with_dev_dependencies(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write(
                '[project]\ndependencies=["foo"]\n[tool.pdm]\nversion = {source ='
                ' "scm"}\n[tool.pdm.dev-dependencies]\ngroup=["bar"]'
            )

        spec = DependencyGetterBuilder(pyproject_toml_path).build()
        assert isinstance(spec, PDMDependencyGetter)


def test_pdm_without_dev_dependencies(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write('[project]\ndependencies=["foo"]\n[tool.pdm]\nversion = {source = "scm"}')

        spec = DependencyGetterBuilder(pyproject_toml_path).build()
        assert isinstance(spec, PEP621DependencyGetter)


def test_pep_621(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write('[project]\ndependencies=["foo"]')

        spec = DependencyGetterBuilder(pyproject_toml_path).build()
        assert isinstance(spec, PEP621DependencyGetter)


def test_both(tmp_path: Path) -> None:
    """
    If both are found, result should be 'poetry'
    """

    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write('[tool.poetry.dependencies]\nfake = "10"')

        with Path("requirements.txt").open("w") as f:
            f.write('foo >= "1.0"')

        spec = DependencyGetterBuilder(pyproject_toml_path).build()
        assert isinstance(spec, PoetryDependencyGetter)


def test_requirements_files(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        with Path("requirements.txt").open("w") as f:
            f.write('foo >= "1.0"')

        spec = DependencyGetterBuilder(Path("pyproject.toml"), requirements_files=("requirements.txt",)).build()
        assert isinstance(spec, RequirementsTxtDependencyGetter)


def test_requirements_files_with_argument_not_root_directory(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        Path("req").mkdir()

        with Path("req/req.txt").open("w") as f:
            f.write('foo >= "1.0"')

        spec = DependencyGetterBuilder(Path("pyproject.toml"), requirements_files=("req/req.txt",)).build()
        assert isinstance(spec, RequirementsTxtDependencyGetter)


def test_dependency_specification_not_found_raises_exception(tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write('[build-system]\nrequires = ["maturin>=1.5,<2.0"]\nbuild-backend = "maturin"')

    with caplog.at_level(logging.DEBUG), run_within_dir(tmp_path), pytest.raises(
        DependencySpecificationNotFoundError,
        match=re.escape(
            "No file called 'pyproject.toml' with a [tool.poetry.dependencies], [tool.pdm] or [project] section or"
            " file(s) called 'req/req.txt' found. Exiting."
        ),
    ):
        DependencyGetterBuilder(Path("pyproject.toml"), requirements_files=("req/req.txt",)).build()

    assert caplog.messages == [
        "pyproject.toml found!",
        (
            "pyproject.toml does not contain a [tool.poetry] section, so Poetry is not used to specify the"
            " project's dependencies."
        ),
        (
            "pyproject.toml does not contain a [tool.pdm.dev-dependencies] section, so PDM is not used to specify the"
            " project's dependencies."
        ),
        (
            "pyproject.toml does not contain a [project] section, so PEP 621 is not used to specify the project's"
            " dependencies."
        ),
    ]


def test_check_for_requirements_in_file_with_requirements_in(tmp_path: Path, caplog: LogCaptureFixture) -> None:
    with run_within_dir(tmp_path):
        # Setup: Create a requirements.in file in the temporary directory
        requirements_in_path = Path("requirements.in")
        requirements_in_path.touch()
        requirements_txt_path = Path("requirements.txt")
        requirements_txt_path.touch()

        # Use caplog to capture logging at the INFO level
        with caplog.at_level(logging.INFO):
            spec = DependencyGetterBuilder(
                config=Path("pyproject.toml"),
                requirements_files=("requirements.txt",),
                using_default_requirements_files=True,
            ).build()

        # Assert that requirements_files is updated correctly
        assert spec.requirements_files == ("requirements.in",)  # type: ignore[attr-defined]

        # Assert that the expected log message is present
        expected_log = (
            "Detected a 'requirements.in' file in the project and no 'requirements-files' were explicitly specified. "
            "Automatically using 'requirements.in' as the source for the project's dependencies. To specify a different source for "
            "the project's dependencies, use the '--requirements-files' option."
        )
        assert expected_log in caplog.text



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from deptry.violations.base import Violation

if TYPE_CHECKING:
    from deptry.dependency import Dependency


@dataclass
class DEP005StandardLibraryDependencyViolation(Violation):
    error_code: ClassVar[str] = "DEP005"
    error_template: ClassVar[str] = (
        "'{name}' is defined as a dependency but it is included in the Python standard library."
    )
    issue: Dependency

    def get_error_message(self) -> str:
        return self.error_template.format(name=self.issue.name)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from deptry.imports.location import Location
from deptry.module import ModuleBuilder, ModuleLocations
from deptry.violations import DEP003TransitiveDependenciesFinder
from deptry.violations.dep003_transitive.violation import DEP003TransitiveDependencyViolation

if TYPE_CHECKING:
    from deptry.dependency import Dependency


def test_simple() -> None:
    """
    black is in testing environment which requires platformdirs, so platformdirs should be found as transitive.
    """
    dependencies: list[Dependency] = []

    module_platformdirs_locations = [Location(Path("foo.py"), 1, 2), Location(Path("bar.py"), 3, 4)]
    module_platformdirs = ModuleBuilder("platformdirs", {"foo"}, frozenset(), dependencies).build()

    modules_locations = [ModuleLocations(module_platformdirs, module_platformdirs_locations)]

    assert DEP003TransitiveDependenciesFinder(modules_locations, dependencies, frozenset()).find() == [
        DEP003TransitiveDependencyViolation(module_platformdirs, location) for location in module_platformdirs_locations
    ]


def test_simple_with_ignore() -> None:
    dependencies: list[Dependency] = []
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("foobar", {"foo"}, frozenset(), dependencies).build(), [Location(Path("foo.py"), 1, 2)]
        )
    ]

    assert (
        DEP003TransitiveDependenciesFinder(
            modules_locations, dependencies, frozenset(), ignored_modules=("foobar",)
        ).find()
        == []
    )


def test_simple_with_standard_library() -> None:
    dependencies: list[Dependency] = []
    standard_library_modules = frozenset(["foobar"])
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("foobar", {"foo"}, standard_library_modules, dependencies).build(),
            [Location(Path("foo.py"), 1, 2)],
        )
    ]

    assert DEP003TransitiveDependenciesFinder(modules_locations, dependencies, frozenset()).find() == []



from __future__ import annotations

import uuid
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from tests.functional.utils import Project
from tests.utils import get_issues_report

if TYPE_CHECKING:
    from tests.utils import PipVenvFactory


@pytest.mark.xdist_group(name=Project.MULTIPLE_SOURCE_DIRECTORIES)
def test_cli_with_multiple_source_directories(pip_venv_factory: PipVenvFactory) -> None:
    with pip_venv_factory(Project.MULTIPLE_SOURCE_DIRECTORIES) as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry src worker -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {"code": "DEP002", "message": "'toml' defined as a dependency but not used in the codebase"},
                "module": "toml",
                "location": {"file": str(Path("pyproject.toml")), "line": None, "column": None},
            },
            {
                "error": {"code": "DEP001", "message": "'httpx' imported but missing from the dependency definitions"},
                "module": "httpx",
                "location": {"file": str(Path("src/foo.py")), "line": 1, "column": 8},
            },
            {
                "error": {"code": "DEP001", "message": "'celery' imported but missing from the dependency definitions"},
                "module": "celery",
                "location": {"file": str(Path("worker/foo.py")), "line": 1, "column": 8},
            },
        ]



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from deptry.reporters.base import Reporter

if TYPE_CHECKING:
    from deptry.imports.location import Location
    from deptry.violations import Violation


COLORS = {
    "BOLD": "\033[1m",
    "CYAN": "\033[36m",
    "GREEN": "\033[32m",
    "RED": "\033[31m",
    "RESET": "\033[m",
}
COLORS_NOOP = {color: "" for color in COLORS}


@dataclass
class TextReporter(Reporter):
    use_ansi: bool = True

    def report(self) -> None:
        self._log_and_exit()

    def _log_and_exit(self) -> None:
        self._log_violations(self.violations)

        self._log_total_number_of_violations_found(self.violations)

    def _log_total_number_of_violations_found(self, violations: list[Violation]) -> None:
        if violations:
            logging.info(
                self._stylize(
                    "{BOLD}{RED}Found {total} dependency {issue_word}.{RESET}",
                    total=len(violations),
                    issue_word="issues" if len(violations) > 1 else "issue",
                )
            )
            logging.info("\nFor more information, see the documentation: https://deptry.com/")
        else:
            logging.info(self._stylize("{BOLD}{GREEN}Success! No dependency issues found.{RESET}"))

    def _log_violations(self, violations: list[Violation]) -> None:
        logging.info("")

        for violation in violations:
            logging.info(self._format_error(violation))

    def _format_error(self, violation: Violation) -> str:
        return self._stylize(
            "{location}{CYAN}:{RESET} {BOLD}{RED}{error_code}{RESET} {error_message}",
            location=self._format_location(violation.location),
            error_code=violation.error_code,
            error_message=violation.get_error_message(),
        )

    def _format_location(self, location: Location) -> str:
        if location.line is not None and location.column is not None:
            return self._stylize(
                "{BOLD}{file}{RESET}{CYAN}:{RESET}{line}{CYAN}:{RESET}{column}",
                file=location.file,
                line=location.line,
                column=location.column,
            )
        return self._stylize("{BOLD}{file}{RESET}", file=location.file)

    def _stylize(self, text: str, **kwargs: Any) -> str:
        return text.format(**kwargs, **self._get_colors())

    def _get_colors(self) -> dict[str, str]:
        return COLORS if self.use_ansi else COLORS_NOOP



from __future__ import annotations

REQUIREMENTS_TXT_DEPRECATION_MESSAGE = (
    "Warning: In an upcoming release, support for the `--requirements-txt` command-line "
    "option and the `requirements_txt` configuration parameter will be discontinued. "
    "Instead, use `--requirements-files` or `requirements_files` under the `[tool.deptry]` "
    "section in pyproject.toml."
)

REQUIREMENTS_TXT_DEV_DEPRECATION_MESSAGE = (
    "Warning: In an upcoming release, support for the `--requirements-txt-dev` command-line "
    "option and the `requirements_txt_dev` configuration parameter will be discontinued. "
    "Instead, use `--requirements-files-dev` or `requirements_files_dev` under the `[tool.deptry]` "
    "section in pyproject.toml."
)



from __future__ import annotations

import contextlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from deptry.dependency import Dependency
from deptry.dependency_getter.base import DependenciesExtract, DependencyGetter
from deptry.utils import load_pyproject_toml

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


@dataclass
class PoetryDependencyGetter(DependencyGetter):
    """Extract Poetry dependencies from pyproject.toml."""

    def get(self) -> DependenciesExtract:
        return DependenciesExtract(self._get_poetry_dependencies(), self._get_poetry_dev_dependencies())

    def _get_poetry_dependencies(self) -> list[Dependency]:
        pyproject_data = load_pyproject_toml(self.config)
        dependencies: dict[str, Any] = pyproject_data["tool"]["poetry"].get("dependencies", {})
        return self._get_dependencies(dependencies, self.package_module_name_map)

    def _get_poetry_dev_dependencies(self) -> list[Dependency]:
        """
        Poetry's development dependencies can be specified under either of the following:

        [tool.poetry.dev-dependencies]
        [tool.poetry.group.dev.dependencies]

        or both.
        """
        dependencies: dict[str, str] = {}
        pyproject_data = load_pyproject_toml(self.config)

        with contextlib.suppress(KeyError):
            dependencies = {**dependencies, **pyproject_data["tool"]["poetry"]["dev-dependencies"]}

        try:
            dependency_groups = pyproject_data["tool"]["poetry"]["group"]
        except KeyError:
            dependency_groups = {}

        for group_values in dependency_groups.values():
            with contextlib.suppress(KeyError):
                dependencies = {**dependencies, **group_values["dependencies"]}

        return self._get_dependencies(dependencies, self.package_module_name_map)

    def _get_dependencies(
        self, poetry_dependencies: dict[str, Any], package_module_name_map: Mapping[str, Sequence[str]]
    ) -> list[Dependency]:
        return [
            Dependency(dep, self.config, module_names=package_module_name_map.get(dep))
            for dep in poetry_dependencies
            if dep != "python"
        ]



from __future__ import annotations

import shlex
import subprocess
import sys
from pathlib import Path

import pytest
import xdist

from tests.functional.utils import DEPTRY_WHEEL_DIRECTORY
from tests.utils import PDMVenvFactory, PipVenvFactory, PoetryVenvFactory


def pytest_sessionstart(session: pytest.Session) -> None:
    # When running the tests on multiple workers with pytest-xdist, the hook will be run several times:
    # - once from "master" node, when test suite starts
    # - X times (where X is the number of workers), when tests start in each worker
    # We only want to run the hook once, so we explicitly skip the hook if running on a pytest-xdist worker.
    if xdist.is_xdist_worker(session):
        return None

    deptry_wheel_path = Path(DEPTRY_WHEEL_DIRECTORY)

    print(f"Building `deptry` wheel in {deptry_wheel_path} to use it on functional tests...")  # noqa: T201

    try:
        result = subprocess.run(
            shlex.split(f"pdm build -v --no-sdist --dest {deptry_wheel_path}", posix=sys.platform != "win32"),
            capture_output=True,
            text=True,
            check=True,
        )
        print("pdm build output: %s", result.stdout)  # noqa: T201
        print("pdm build errors: %s", result.stderr)  # noqa: T201
    except subprocess.CalledProcessError as e:
        print("Output: %s", e.output)  # noqa: T201
        print("Errors: %s", e.stderr)  # noqa: T201
        raise


@pytest.fixture(scope="session")
def pdm_venv_factory(tmp_path_factory: pytest.TempPathFactory) -> PDMVenvFactory:
    return PDMVenvFactory(tmp_path_factory.getbasetemp() / "venvs")


@pytest.fixture(scope="session")
def poetry_venv_factory(tmp_path_factory: pytest.TempPathFactory) -> PoetryVenvFactory:
    return PoetryVenvFactory(tmp_path_factory.getbasetemp() / "venvs")


@pytest.fixture(scope="session")
def pip_venv_factory(tmp_path_factory: pytest.TempPathFactory) -> PipVenvFactory:
    return PipVenvFactory(tmp_path_factory.getbasetemp() / "venvs")



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.imports.location import Location
from deptry.violations.base import ViolationsFinder
from deptry.violations.dep005_standard_library.violation import DEP005StandardLibraryDependencyViolation

if TYPE_CHECKING:
    from deptry.violations import Violation


@dataclass
class DEP005StandardLibraryDependenciesFinder(ViolationsFinder):
    """
    Finds dependencies that are part of the standard library but are defined as dependencies.
    """

    violation = DEP005StandardLibraryDependencyViolation

    def find(self) -> list[Violation]:
        logging.debug("\nScanning for dependencies that are part of the standard library...")
        stdlib_violations: list[Violation] = []

        for dependency in self.dependencies:
            logging.debug("Scanning module %s...", dependency.name)

            if dependency.name in self.standard_library_modules:
                if dependency.name in self.ignored_modules:
                    logging.debug(
                        "Dependency '%s' found to be a dependency that is part of the standard library, but ignoring.",
                        dependency.name,
                    )
                    continue

                logging.debug(
                    "Dependency '%s' marked as a dependency that is part of the standard library.", dependency.name
                )
                stdlib_violations.append(self.violation(dependency, Location(dependency.definition_file)))

        return stdlib_violations



from __future__ import annotations

import operator
from typing import TYPE_CHECKING

from deptry.violations import (
    DEP001MissingDependenciesFinder,
    DEP002UnusedDependenciesFinder,
    DEP003TransitiveDependenciesFinder,
    DEP004MisplacedDevDependenciesFinder,
    DEP005StandardLibraryDependenciesFinder,
)

if TYPE_CHECKING:
    from typing import Mapping

    from deptry.dependency import Dependency
    from deptry.module import ModuleLocations
    from deptry.violations import Violation, ViolationsFinder


_VIOLATIONS_FINDERS: tuple[type[ViolationsFinder], ...] = (
    DEP001MissingDependenciesFinder,
    DEP002UnusedDependenciesFinder,
    DEP003TransitiveDependenciesFinder,
    DEP004MisplacedDevDependenciesFinder,
    DEP005StandardLibraryDependenciesFinder,
)


def find_violations(
    imported_modules_with_locations: list[ModuleLocations],
    dependencies: list[Dependency],
    ignore: tuple[str, ...],
    per_rule_ignores: Mapping[str, tuple[str, ...]],
    standard_library_modules: frozenset[str],
) -> list[Violation]:
    violations = []

    for violation_finder in _VIOLATIONS_FINDERS:
        if violation_finder.violation.error_code not in ignore:
            violations.extend(
                violation_finder(
                    imported_modules_with_locations=imported_modules_with_locations,
                    dependencies=dependencies,
                    ignored_modules=per_rule_ignores.get(violation_finder.violation.error_code, ()),
                    standard_library_modules=standard_library_modules,
                ).find()
            )
    return _get_sorted_violations(violations)


def _get_sorted_violations(violations: list[Violation]) -> list[Violation]:
    return sorted(
        violations, key=operator.attrgetter("location.file", "location.line", "location.column", "error_code")
    )



from __future__ import annotations

import uuid
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from tests.functional.utils import Project
from tests.utils import get_issues_report

if TYPE_CHECKING:
    from tests.utils import PipVenvFactory


@pytest.mark.xdist_group(name=Project.REQUIREMENTS_IN)
def test_cli_single_requirements_files(pip_venv_factory: PipVenvFactory) -> None:
    """
    in this case, deptry should recognize that there is a `requirements.in` in the project, and
    use that as the source of the dependencies.
    """
    with pip_venv_factory(
        Project.REQUIREMENTS_IN,
        install_command=("pip install -r requirements.txt -r requirements-dev.txt"),
    ) as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {"code": "DEP002", "message": "'isort' defined as a dependency but not used in the codebase"},
                "module": "isort",
                "location": {"file": str(Path("requirements.in")), "line": None, "column": None},
            },
            {
                "error": {"code": "DEP002", "message": "'pandas' defined as a dependency but not used in the codebase"},
                "module": "pandas",
                "location": {"file": str(Path("requirements.in")), "line": None, "column": None},
            },
            {
                "error": {"code": "DEP004", "message": "'black' imported but declared as a dev dependency"},
                "module": "black",
                "location": {"file": str(Path("src/main.py")), "line": 4, "column": 8},
            },
            {
                "error": {"code": "DEP001", "message": "'white' imported but missing from the dependency definitions"},
                "module": "white",
                "location": {"file": str(Path("src/main.py")), "line": 6, "column": 8},
            },
            {
                "error": {"code": "DEP003", "message": "'requests' imported but it is a transitive dependency"},
                "module": "requests",
                "location": {"file": str(Path("src/main.py")), "line": 8, "column": 8},
            },
        ]


@pytest.mark.xdist_group(name=Project.REQUIREMENTS_IN)
def test_cli_multiple_requirements_files(pip_venv_factory: PipVenvFactory) -> None:
    """
    in this case, deptry recognizes that there is a `requirements.in` in the project, but the user
    can overwrite that with '--requirements-files requirements.txt', so it still takes requirements.txt as the source
    for the project's dependencies.
    """
    with pip_venv_factory(
        Project.REQUIREMENTS_IN,
        install_command=("pip install -r requirements.txt -r requirements-dev.txt"),
    ) as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . --requirements-files requirements.txt -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {"code": "DEP002", "message": "'isort' defined as a dependency but not used in the codebase"},
                "module": "isort",
                "location": {"file": str(Path("requirements.txt")), "line": None, "column": None},
            },
            {
                "error": {"code": "DEP002", "message": "'pandas' defined as a dependency but not used in the codebase"},
                "module": "pandas",
                "location": {"file": str(Path("requirements.txt")), "line": None, "column": None},
            },
            {
                "error": {"code": "DEP002", "message": "'numpy' defined as a dependency but not used in the codebase"},
                "module": "numpy",
                "location": {"file": str(Path("requirements.txt")), "line": None, "column": None},
            },
            {
                "error": {"code": "DEP004", "message": "'black' imported but declared as a dev dependency"},
                "module": "black",
                "location": {"file": str(Path("src/main.py")), "line": 4, "column": 8},
            },
            {
                "error": {"code": "DEP001", "message": "'white' imported but missing from the dependency definitions"},
                "module": "white",
                "location": {"file": str(Path("src/main.py")), "line": 6, "column": 8},
            },
        ]



from __future__ import annotations

import uuid
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from tests.functional.utils import Project
from tests.utils import get_issues_report

if TYPE_CHECKING:
    from tests.utils import PipVenvFactory


@pytest.mark.xdist_group(name=Project.PYPROJECT_DIFFERENT_DIRECTORY)
def test_cli_with_pyproject_different_directory(pip_venv_factory: PipVenvFactory) -> None:
    with pip_venv_factory(
        Project.PYPROJECT_DIFFERENT_DIRECTORY, install_command="pip install ./a_sub_directory"
    ) as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry src --config a_sub_directory/pyproject.toml -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("a_sub_directory/pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("a_sub_directory/pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'mypy' defined as a dependency but not used in the codebase",
                },
                "module": "mypy",
                "location": {
                    "file": str(Path("a_sub_directory/pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'pytest' defined as a dependency but not used in the codebase",
                },
                "module": "pytest",
                "location": {
                    "file": str(Path("a_sub_directory/pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP001",
                    "message": "'white' imported but missing from the dependency definitions",
                },
                "module": "white",
                "location": {
                    "file": str(Path("src/project_with_src_directory/foo.py")),
                    "line": 6,
                    "column": 8,
                },
            },
        ]



from __future__ import annotations

from unittest.mock import ANY, patch

from click.testing import CliRunner

from deptry.cli import deptry
from deptry.deprecate.requirements_txt import (
    REQUIREMENTS_TXT_DEPRECATION_MESSAGE,
    REQUIREMENTS_TXT_DEV_DEPRECATION_MESSAGE,
)

DEFAULT_CORE_ARGS = {
    "root": (ANY,),
    "config": ANY,
    "no_ansi": ANY,
    "exclude": ANY,
    "extend_exclude": ANY,
    "using_default_exclude": ANY,
    "ignore_notebooks": ANY,
    "ignore": ANY,
    "per_rule_ignores": ANY,
    "known_first_party": ANY,
    "json_output": ANY,
    "package_module_name_map": ANY,
    "pep621_dev_dependency_groups": ANY,
    "using_default_requirements_files": ANY,
}


def test_requirements_txt_deprecated() -> None:
    with patch("deptry.cli.Core") as mock_core, patch("logging.warning") as mock_warning:
        result = CliRunner().invoke(deptry, [".", "--requirements-txt", "somefile.txt"])

        assert result.exit_code == 0
        mock_warning.assert_called_once_with(REQUIREMENTS_TXT_DEPRECATION_MESSAGE)

        # Assert that Core was instantiated with the correct arguments
        mock_core.assert_called_once_with(
            **DEFAULT_CORE_ARGS,
            requirements_files=("somefile.txt",),
            requirements_files_dev=("dev-requirements.txt", "requirements-dev.txt"),
        )


def test_requirements_txt_dev_deprecated() -> None:
    with patch("deptry.cli.Core") as mock_core, patch("logging.warning") as mock_warning:
        result = CliRunner().invoke(deptry, [".", "--requirements-txt-dev", "somefile.txt"])

        assert result.exit_code == 0
        mock_warning.assert_called_once_with(REQUIREMENTS_TXT_DEV_DEPRECATION_MESSAGE)

        # Assert that Core was instantiated with the correct arguments
        mock_core.assert_called_once_with(
            **DEFAULT_CORE_ARGS, requirements_files=("requirements.txt",), requirements_files_dev=("somefile.txt",)
        )


def test_requirements_files_works_as_expected() -> None:
    with patch("deptry.cli.Core") as mock_core, patch("logging.warning") as mock_warning:
        result = CliRunner().invoke(deptry, [".", "--requirements-files", "somefile.txt"])

        assert result.exit_code == 0
        mock_warning.assert_not_called()

        # Assert that Core was instantiated with the correct arguments
        mock_core.assert_called_once_with(
            **DEFAULT_CORE_ARGS,
            requirements_files=("somefile.txt",),
            requirements_files_dev=("dev-requirements.txt", "requirements-dev.txt"),
        )


def test_requirements_files_dev_works_as_expected() -> None:
    with patch("deptry.cli.Core") as mock_core, patch("logging.warning") as mock_warning:
        result = CliRunner().invoke(deptry, [".", "--requirements-files-dev", "somefile.txt"])

        assert result.exit_code == 0
        mock_warning.assert_not_called()

        # Assert that Core was instantiated with the correct arguments
        mock_core.assert_called_once_with(
            **DEFAULT_CORE_ARGS, requirements_files=("requirements.txt",), requirements_files_dev=("somefile.txt",)
        )



from __future__ import annotations

import logging
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.dependency_getter.builder import DependencyGetterBuilder
from deptry.exceptions import UnsupportedPythonVersionError
from deptry.imports.extract import get_imported_modules_from_list_of_files
from deptry.module import ModuleBuilder, ModuleLocations
from deptry.python_file_finder import get_all_python_files_in
from deptry.reporters import JSONReporter, TextReporter
from deptry.stdlibs import STDLIBS_PYTHON
from deptry.violations.finder import find_violations

if TYPE_CHECKING:
    from collections.abc import Mapping
    from pathlib import Path

    from deptry.dependency_getter.base import DependenciesExtract
    from deptry.violations import Violation


@dataclass
class Core:
    root: tuple[Path, ...]
    config: Path
    no_ansi: bool
    per_rule_ignores: Mapping[str, tuple[str, ...]]
    ignore: tuple[str, ...]
    exclude: tuple[str, ...]
    extend_exclude: tuple[str, ...]
    using_default_exclude: bool
    ignore_notebooks: bool
    requirements_files: tuple[str, ...]
    using_default_requirements_files: bool
    requirements_files_dev: tuple[str, ...]
    known_first_party: tuple[str, ...]
    json_output: str
    package_module_name_map: Mapping[str, tuple[str, ...]]
    pep621_dev_dependency_groups: tuple[str, ...]

    def run(self) -> None:
        self._log_config()

        dependency_getter = DependencyGetterBuilder(
            self.config,
            self.package_module_name_map,
            self.pep621_dev_dependency_groups,
            self.requirements_files,
            self.using_default_requirements_files,
            self.requirements_files_dev,
        ).build()

        dependencies_extract = dependency_getter.get()

        self._log_dependencies(dependencies_extract)

        python_files = self._find_python_files()
        local_modules = self._get_local_modules()
        standard_library_modules = self._get_standard_library_modules()

        imported_modules_with_locations = [
            ModuleLocations(
                ModuleBuilder(
                    module,
                    local_modules,
                    standard_library_modules,
                    dependencies_extract.dependencies,
                    dependencies_extract.dev_dependencies,
                ).build(),
                locations,
            )
            for module, locations in get_imported_modules_from_list_of_files(python_files).items()
        ]

        violations = find_violations(
            imported_modules_with_locations,
            dependencies_extract.dependencies,
            self.ignore,
            self.per_rule_ignores,
            standard_library_modules,
        )
        TextReporter(violations, use_ansi=not self.no_ansi).report()

        if self.json_output:
            JSONReporter(violations, self.json_output).report()

        self._exit(violations)

    def _find_python_files(self) -> list[Path]:
        logging.debug("Collecting Python files to scan...")

        python_files = get_all_python_files_in(
            self.root, self.exclude, self.extend_exclude, self.using_default_exclude, self.ignore_notebooks
        )

        logging.debug(
            "Python files to scan for imports:\n%s\n", "\n".join(str(python_file) for python_file in python_files)
        )

        return python_files

    def _get_local_modules(self) -> set[str]:
        """
        Get all local Python modules from the source directories and `known_first_party` list.
        A module is considered a local Python module if it matches at least one of those conditions:
        - it is a directory that contains at least one Python file
        - it is a Python file that is not named `__init__.py` (since it is a special case)
        - it is set in the `known_first_party` list
        """
        guessed_local_modules = {
            path.stem for source in self.root for path in source.iterdir() if self._is_local_module(path)
        }

        return guessed_local_modules | set(self.known_first_party)

    @staticmethod
    def _is_local_module(path: Path) -> bool:
        """Guess if a module is a local Python module."""
        return bool(
            (path.is_file() and path.name != "__init__.py" and path.suffix == ".py")
            or (path.is_dir() and list(path.rglob("*.py")))
        )

    @staticmethod
    def _get_standard_library_modules() -> frozenset[str]:
        if sys.version_info[:2] >= (3, 10):
            return sys.stdlib_module_names

        try:  # type: ignore[unreachable, unused-ignore]
            return STDLIBS_PYTHON[f"{sys.version_info[0]}{sys.version_info[1]}"]
        except KeyError as e:
            raise UnsupportedPythonVersionError((sys.version_info[0], sys.version_info[1])) from e

    def _log_config(self) -> None:
        logging.debug("Running with the following configuration:")
        for key, value in vars(self).items():
            logging.debug("%s: %s", key, value)
        logging.debug("")

    @staticmethod
    def _log_dependencies(dependencies_extract: DependenciesExtract) -> None:
        if dependencies_extract.dependencies:
            logging.debug("The project contains the following dependencies:")
            for dependency in dependencies_extract.dependencies:
                logging.debug(dependency)
            logging.debug("")

        if dependencies_extract.dev_dependencies:
            logging.debug("The project contains the following dev dependencies:")
            for dependency in dependencies_extract.dev_dependencies:
                logging.debug(dependency)
            logging.debug("")

    @staticmethod
    def _exit(violations: list[Violation]) -> None:
        sys.exit(bool(violations))



from __future__ import annotations

from pathlib import Path

from deptry.dependency import Dependency
from deptry.imports.location import Location
from deptry.module import ModuleBuilder, ModuleLocations
from deptry.violations import DEP005StandardLibraryDependenciesFinder, DEP005StandardLibraryDependencyViolation


def test_simple() -> None:
    dependency_asyncio = Dependency("asyncio", Path("pyproject.toml"))
    dependencies = [dependency_asyncio]
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("asyncio", {"foo"}, frozenset(), dependencies).build(), [Location(Path("foo.py"), 1, 2)]
        )
    ]

    assert DEP005StandardLibraryDependenciesFinder(
        imported_modules_with_locations=modules_locations,
        dependencies=dependencies,
        standard_library_modules=frozenset(["asyncio"]),
    ).find() == [DEP005StandardLibraryDependencyViolation(dependency_asyncio, Location(Path("pyproject.toml")))]


def test_simple_with_ignore() -> None:
    dependency_asyncio = Dependency("asyncio", Path("pyproject.toml"))
    dependencies = [dependency_asyncio]
    modules_locations = [
        ModuleLocations(
            ModuleBuilder("asyncio", {"foo"}, frozenset(), dependencies).build(), [Location(Path("foo.py"), 1, 2)]
        )
    ]

    assert (
        DEP005StandardLibraryDependenciesFinder(
            imported_modules_with_locations=modules_locations,
            dependencies=dependencies,
            standard_library_modules=frozenset(["asyncio"]),
            ignored_modules=("asyncio",),
        ).find()
        == []
    )



from __future__ import annotations

from pathlib import Path

from deptry.imports.location import Location
from deptry.module import Module
from deptry.violations import DEP001MissingDependencyViolation, DEP004MisplacedDevDependencyViolation
from deptry.violations.finder import _get_sorted_violations


def test__get_sorted_violations() -> None:
    violations = [
        DEP004MisplacedDevDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 2, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 3, 1)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 2, 1)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 3, 0)),
    ]

    assert _get_sorted_violations(violations) == [
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 2, 1)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 3, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 3, 1)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 0)),
        DEP004MisplacedDevDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 2, 0)),
    ]
